package gregapi.config;

import static gregapi.data.CS.GAPI;
import static gregapi.data.CS.GT;
import static gregapi.data.CS.T;
import gregapi.GT_API;
import gregapi.data.CS.DirectoriesGT;
import gregapi.util.UT;

import java.io.File;

import net.minecraft.item.ItemStack;
import net.minecraftforge.common.config.Configuration;
import net.minecraftforge.common.config.Property;

/**
 * @author Gregorius Techneticies
 */
public class Config implements Runnable {
	public static Configuration sConfigFileIDs;
	private static boolean sIDConfigNeedsSaving;
	private boolean mUsesDefaultsInNames = T;
	
	public static int addIDConfig(Object aCategory, String aName, int aDefault) {
		if (UT.Code.stringInvalid(aName)) return aDefault;
		Property tProperty = sConfigFileIDs.get(aCategory.toString().replaceAll("\\|", "."), aName.replaceAll("\\|", "."), aDefault);
		int rResult = tProperty.getInt(aDefault);
		if (!tProperty.wasRead()) if (GT_API.sEnableDynamicWritingToConfig) sConfigFileIDs.save(); else sIDConfigNeedsSaving = true;
		return rResult;
	}
	
	public final Configuration mConfig;
	
	public Config(String aFileName) {this(new File(DirectoriesGT.CONFIG_GT, aFileName));}
	public Config(String aPath, String aFileName) {this(new File(aPath, aFileName));}
	public Config(File aPath, String aFileName) {this(new File(aPath, aFileName));}
	public Config(File aConfig) {this(new Configuration(aConfig));}
	
	public Config(Configuration aConfig) {
		mConfig = aConfig;
		mConfig.load();
		mConfig.save();
		if (GAPI.mAfterPreInit		!= null) GAPI.mAfterPreInit		.add(this); // in case of crash on startup
		if (GAPI.mAfterInit			!= null) GAPI.mAfterInit		.add(this); // in case of crash on startup
		if (GAPI.mAfterPostInit		!= null) GAPI.mAfterPostInit	.add(this); // in case of the Player not starting a Game
		if (GT != null) {
		if (GT .mAfterInit			!= null) GT .mAfterInit			.add(this); // in case of crash on startup
		if (GT .mAfterPreInit		!= null) GT .mAfterPreInit		.add(this); // in case of crash on startup
		if (GT .mAfterPostInit		!= null) GT .mAfterPostInit		.add(this); // in case of the Player not starting a Game
		}
	}
	
	@Override
	public void run() {
		mConfig.save();
		if (sIDConfigNeedsSaving) {
			sConfigFileIDs.save();
			sIDConfigNeedsSaving = false;
		}
	}
	
	public Config setUseDefaultInNames(boolean aUsesDefaultsInNames) {
		mUsesDefaultsInNames = aUsesDefaultsInNames;
		return this;
	}
	
	public boolean get(Object aCategory, ItemStack aStack, boolean aDefault) {
		return get(aCategory, UT.Stacks.configName(aStack), aDefault);
	}
	
	public boolean get(Object aCategory, String aName, boolean aDefault) {
		if (UT.Code.stringInvalid(aName)) return aDefault;
		Property tProperty = mConfig.get(aCategory.toString().replaceAll("\\|", "_"), (aName+(mUsesDefaultsInNames?"_"+aDefault:"")).replaceAll("\\|", "_"), aDefault);
		boolean rResult = tProperty.getBoolean(aDefault);
		if (GT_API.sEnableDynamicWritingToConfig && !tProperty.wasRead()) mConfig.save();
		return rResult;
	}
	
	public int get(Object aCategory, ItemStack aStack, long aDefault) {
		return get(aCategory, UT.Stacks.configName(aStack), aDefault);
	}
	
	public int get(Object aCategory, String aName, long aDefault) {
		if (UT.Code.stringInvalid(aName)) return UT.Code.bindInt(aDefault);
		Property tProperty = mConfig.get(aCategory.toString().replaceAll("\\|", "_"), (aName+(mUsesDefaultsInNames?"_"+UT.Code.bindInt(aDefault):"")).replaceAll("\\|", "_"), UT.Code.bindInt(aDefault));
		int rResult = tProperty.getInt(UT.Code.bindInt(aDefault));
		if (GT_API.sEnableDynamicWritingToConfig && !tProperty.wasRead()) mConfig.save();
		return rResult;
	}
	
	public String get(Object aCategory, ItemStack aStack, String aDefault) {
		return get(aCategory, UT.Stacks.configName(aStack), aDefault);
	}
	
	public String get(Object aCategory, String aName, String aDefault) {
		if (UT.Code.stringInvalid(aName)) return aDefault;
		Property tProperty = mConfig.get(aCategory.toString().replaceAll("\\|", "_"), (aName+(mUsesDefaultsInNames?"_"+aDefault:"")).replaceAll("\\|", "_"), aDefault);
		String rResult = tProperty.getString();
		if (GT_API.sEnableDynamicWritingToConfig && !tProperty.wasRead()) mConfig.save();
		return rResult;
	}
	
	public double get(Object aCategory, ItemStack aStack, double aDefault) {
		return get(aCategory, UT.Stacks.configName(aStack), aDefault);
	}
	
	public double get(Object aCategory, String aName, double aDefault) {
		if (UT.Code.stringInvalid(aName)) return aDefault;
		Property tProperty = mConfig.get(aCategory.toString().replaceAll("\\|", "_"), (aName+(mUsesDefaultsInNames?"_"+aDefault:"")).replaceAll("\\|", "_"), aDefault);
		double rResult = tProperty.getDouble(aDefault);
		if (GT_API.sEnableDynamicWritingToConfig && !tProperty.wasRead()) mConfig.save();
		return rResult;
	}
	
	@Override public String toString() {return "Config File";}
}