package gregapi.old;

import static gregapi.data.CS.B;
import static gregapi.data.CS.F;
import static gregapi.data.CS.T;
import gregapi.GT_API;
import gregapi.code.ItemStackContainer;
import gregapi.code.ItemStackMap;
import gregapi.code.ItemStackSet;
import gregapi.old.Textures.BlockIcons;
import gregapi.render.BlockTextureDefault;
import gregapi.render.ITexture;
import gregapi.util.UT;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import net.minecraft.block.Block;
import net.minecraft.item.ItemStack;
import net.minecraft.world.World;

/**
 * @author Gregorius Techneticies
 * 
 * Please do not include this File in your Mod-download as it ruins compatiblity, like with the IC2-API
 * You may just copy those Functions into your Code, or better call them via reflection.
 * 
 * The whole API is the basic construct of my Mod. Everything is dependent on it.
 * I change things quite often so please don't include any File inside your Mod, even if it is an Interface.
 * Since some Authors were stupid enough to break this simple Rule, I added Version checks to enforce it.
 * 
 * In these Folders are many useful Functions. You can use them via reflection if you want.
 * I know not everything is compilable due to API's of other Mods, but these are easy to fix in your Setup.
 * 
 * You can use this to learn about Modding, but I would recommend simpler Mods.
 * You may even copypaste Code from these API-Files into your Mod, as I have nothing against that, but you should look exactly at what you are copying.
 */
public class GregTech_API {
	/** Getting assigned by the Config */
	public static boolean sDrinksAlwaysDrinkable = F, sMultiThreadedSounds = F, sColoredGUI = T, sConstantEnergy = T, sMachineExplosions = T, sMachineFlammable = T, sMachineNonWrenchExplosions = T, sMachineRainExplosions = T, sMachineThunderExplosions = T, sMachineFireExplosions = T, sMachineWireFire = true;
	
	/** The Icon List for Covers */
	public static final Map<ItemStackContainer, ITexture> sCovers = new ItemStackMap<ItemStackContainer, ITexture>();
	
	/** The List of Cover Behaviors for the Covers */
	public static final Map<ItemStackContainer, GT_CoverBehavior> sCoverBehaviors = new ItemStackMap<ItemStackContainer, GT_CoverBehavior>();
	
	/** The List of Circuit Behaviors for the Redstone Circuit Block */
	public static final Map<Integer, GT_CircuitryBehavior> sCircuitryBehaviors = new HashMap<Integer, GT_CircuitryBehavior>();
	
	/** The List of Blocks, which can conduct Machine Block Updates */
    public static final Map<Block, Integer> sMachineIDs = new HashMap<Block, Integer>();
    
	/** The Redstone Frequencies */
    public static final Map<Integer, Byte> sWirelessRedstone = new HashMap<Integer, Byte>();
    
	/** The IDSU Frequencies */
	public static final Map<Integer, Integer> sIDSUList = new HashMap<Integer, Integer>();
	
	/** The List of all Sounds used in GT, indices are in the static Block at the bottom */
	public static final Map<Integer, String> sSoundList = new HashMap<Integer, String>();
	
	/** The List of Hazmat Armors */
	public static final ItemStackSet<ItemStackContainer> sGasHazmatList = new ItemStackSet<ItemStackContainer>(), sBioHazmatList = new ItemStackSet<ItemStackContainer>(), sFrostHazmatList = new ItemStackSet<ItemStackContainer>(), sHeatHazmatList = new ItemStackSet<ItemStackContainer>(), sRadioHazmatList = new ItemStackSet<ItemStackContainer>(), sElectroHazmatList = new ItemStackSet<ItemStackContainer>();
	
	/**
	 * Causes a Machineblock Update
	 * This update will cause surrounding MultiBlock Machines to update their Configuration.
	 * You should call this Function in @Block.breakBlock and in @Block.onBlockAdded of your Machine.
	 * @param aWorld is being the World
	 * @param aX is the X-Coord of the update causing Block
	 * @param aY is the Y-Coord of the update causing Block
	 * @param aZ is the Z-Coord of the update causing Block
	 */
	public static boolean causeMachineUpdate(World aWorld, int aX, int aY, int aZ) {
		if (!aWorld.isRemote) new Thread(new GT_Runnable_MachineBlockUpdate(aWorld, aX, aY, aZ), "Machine Block Updating").start();
		return T;
	}
	
	/**
	 * Adds a Multi-Machine Block, like my Machine Casings for example.
	 * You should call @causeMachineUpdate in @Block.breakBlock and in @Block.onBlockAdded of your registered Block.
	 * You don't need to register TileEntities which implement @IMachineBlockUpdateable
	 * @param aID the ID of your Block
	 * @param aMeta the Metadata of the Blocks as Bitmask! -1 or ~0 for all Metavalues
	 */
	public static boolean registerMachineBlock(Block aBlock, int aMeta) {
		if (aBlock == null) return F;
        if (GT_API.sCompatTC != null) GT_API.sCompatTC.registerPortholeBlacklistedBlock(aBlock);
		sMachineIDs.put(aBlock, aMeta);
		return T;
	}
	
	/**
	 * Like above but with boolean Parameters instead of a BitMask
	 */
	public static boolean registerMachineBlock(Block aBlock, boolean... aMeta) {
		if (aBlock == null || aMeta == null || aMeta.length == 0) return F;
        if (GT_API.sCompatTC != null) GT_API.sCompatTC.registerPortholeBlacklistedBlock(aBlock);
		int rMeta = 0;
		for (byte i = 0; i < aMeta.length && i < 16; i++) if (aMeta[i]) rMeta |= B[i];
		sMachineIDs.put(aBlock, rMeta);
		return T;
	}
	
	/**
	 * if this Block is a Machine Update Conducting Block
	 */
	public static boolean isMachineBlock(Block aBlock, int aMeta) {
		if (aBlock == null) return F;
		return (sMachineIDs.containsKey(aBlock) && (sMachineIDs.get(aBlock) & B[aMeta]) != 0);
	}
	
	public static void registerCover(ItemStack aStack, ITexture aCover, GT_CoverBehavior aBehavior) {
		if (!sCovers.containsKey(new ItemStackContainer(aStack))) sCovers.put(new ItemStackContainer(aStack), aCover==null||!aCover.isValidTexture()?new BlockTextureDefault(BlockIcons.RENDERING_ERROR):aCover);
		if (aBehavior != null) sCoverBehaviors.put(new ItemStackContainer(aStack), aBehavior);
	}
	
	public static void registerCoverBehavior(ItemStack aStack, GT_CoverBehavior aBehavior) {
		sCoverBehaviors.put(new ItemStackContainer(aStack), aBehavior==null?sDefaultBehavior:aBehavior);
	}
	
	/**
	 * Registers multiple Cover Items. I use that for the OreDict Functionality.
	 * @param aBehavior can be null
	 */
	public static void registerCover(Collection<ItemStack> aStackList, ITexture aCover, GT_CoverBehavior aBehavior) {
		if (aCover.isValidTexture()) for (ItemStack tStack : aStackList) registerCover(tStack, aCover, aBehavior);
	}
	
	/**
	 * This is the generic Cover behavior. Used for the default Covers, which have no Behavior.
	 */
	public static final GT_CoverBehavior sDefaultBehavior = new GT_Cover_Default(), sNoBehavior = new GT_Cover_None();
	
	/**
	 * returns a Cover behavior, guaranteed to not return null after preload
	 */
	public static GT_CoverBehavior getCoverBehavior(ItemStack aStack) {
		if (aStack == null || aStack.getItem() == null) return sNoBehavior;
		GT_CoverBehavior rCover = sCoverBehaviors.get(new ItemStackContainer(aStack));
		if (rCover == null) return sDefaultBehavior;
		return rCover;
	}
	
	/**
	 * returns a Cover behavior, guaranteed to not return null
	 */
	public static GT_CoverBehavior getCoverBehavior(int aStack) {
		if (aStack == 0) return sNoBehavior;
		return getCoverBehavior(UT.Stacks.toStack(aStack));
	}
}