package gregapi.oredict;

import static gregapi.data.CS.*;
import gregapi.code.*;
import gregapi.config.Config;
import gregapi.data.CS.ModIDs;
import gregapi.data.OP;
import gregapi.data.TD;
import gregapi.oredict.IOreDictListenerEvent.OreDictRegistrationContainer;
import gregapi.oredict.IOreDictRecyclableListener.OreDictRecyclingContainer;
import gregapi.recipes.GT_ModHandler;
import gregapi.util.OM;
import gregapi.util.UT;

import java.util.*;

import net.minecraft.init.Blocks;
import net.minecraft.init.Items;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraftforge.common.MinecraftForge;
import net.minecraftforge.fluids.FluidContainerRegistry;
import net.minecraftforge.fluids.FluidContainerRegistry.FluidContainerData;
import net.minecraftforge.fluids.FluidContainerRegistry.FluidContainerRegisterEvent;
import net.minecraftforge.fluids.IFluidContainerItem;
import net.minecraftforge.oredict.OreDictionary;
import net.minecraftforge.oredict.OreDictionary.OreRegisterEvent;
import cpw.mods.fml.common.FMLLog;
import cpw.mods.fml.common.Loader;
import cpw.mods.fml.common.ModContainer;
import cpw.mods.fml.common.eventhandler.SubscribeEvent;

/**
 * @author Gregorius Techneticies
 * 
 * This Class is for filtering out the Private and Invalid OreDictionary Events and passing the proper ones to any interested User.
 * It also contains a lot of Utility, which was formerly inside the OreDictUnificator.
 */
public final class OreDictManager {
	public static final OreDictManager INSTANCE = new OreDictManager();
	
	/** Those Listeners can also be registered at certain Prefixes in order to only get their Events */
	private final Set<IOreDictListenerEvent> mGlobalOreDictListeners = new HashSetNoNulls<IOreDictListenerEvent>();
	private final Set<OreDictRegistrationContainer> mGlobalRegistrations = new HashSetNoNulls<OreDictRegistrationContainer>();
	
	/** These Listeners always get notified if an Item gets a recyclable ItemData Tag attached to it. */
	private final Set<IOreDictRecyclableListener> mRecyclableOreDictListeners = new HashSetNoNulls<IOreDictRecyclableListener>();
	private final Set<OreDictRecyclingContainer> mRecyclableRegistrations = new HashSetNoNulls<OreDictRecyclingContainer>();
	
	/** Put OreDict Strings which should be registered as something else, right here using addReRegistration */
	private final Map<String, Collection<String>> mReRegistrationMappings = new HashMap<String, Collection<String>>(); 
	/** Assign Item Data to certain OreDict Strings for recycling and similar. Used for plankWood and slabWood for example.*/
	private final Map<String, OreDictItemData> mStringToItemDataMappings = new HashMap<String, OreDictItemData>(); 
	
	/** Lists all Names (with their Prefixes) which don't have a Material assigned to them. */
	private final Set<String> mUnknownMaterials = new HashSetNoNulls<String>();
	/** Lists all Names which don't have a Prefix assigned to them. */
	private final Set<String> mUnknownNames = new HashSetNoNulls<String>();
	/** Lists all Names which get automatically blacklisted. */
	private final Set<String> mAutoBlackListedNames = new HashSetNoNulls<String>();
	/** Lists all Names which are definitely not unknown. */
	private final Set<String> mNotUnknownNames = new HashSetNoNulls<String>();
	/** Lists all Names which should not be processed due to technical Issues. */
	private final Set<String> mIgnoredNames = new HashSetNoNulls<String>(Arrays.asList("plateLapis", "shardEntropy", "shardAir", "shardWater", "shardEarth", "shardFire", "shardOrder", "greggy_greg_do_please_kindly_stuff_a_sock_in_it"));
	/** Lists all Names which have already been registered. Used for the OreDict Listeners, so that certain Recipes (such as Crafting) don't get registered twice.*/
	private final Set<String> mAlreadyRegisteredNames = new HashSetNoNulls<String>();
	/** Used to check if Recipe Outputs accidentally contain uncopied OreDict Items. */
	private final Set<ItemStack> mAllRegisteredOres = new HashSetNoNulls<ItemStack>();
	
	/** The Unification Config File */
	public Config mUnificationConfig;
	
	private boolean mIsRunningInIterationMode = false;
	
	private OreDictManager() {
		mIsRunningInIterationMode = true;
		for (String tOreName : OreDictionary.getOreNames()) for (ItemStack tOreStack : OreDictionary.getOres(tOreName)) onOreRegistration1(new OreRegisterEvent(tOreName, tOreStack));
	    for (FluidContainerData tData : FluidContainerRegistry.getRegisteredFluidContainerData()) onFluidContainerRegistration(new FluidContainerRegisterEvent(tData));
		mIsRunningInIterationMode = false;
		MinecraftForge.EVENT_BUS.register(this);
	}
	
	/**
	 * This Function adds a global Listener for advanced OreDictionary Events.
	 * Unlike the Forge Event System, which replaced Eloraams better System, this Interface ensures, like the great old System, that also all the older Events, which have been called before your Listener-Registration, are getting passed onto your Listener.
	 */
	public void addListener(IOreDictListenerEvent aListener) {
		if (GAPI.mStartedPostInit) addListenerInternal(aListener); else mBufferedListeners1.add(aListener);
	}
	
	private final Set<IOreDictListenerEvent> mBufferedListeners1 = new HashSetNoNulls<IOreDictListenerEvent>();
	private void addListenerInternal(IOreDictListenerEvent aListener) {
		if (mGlobalOreDictListeners.add(aListener)) for (OreDictRegistrationContainer tEvent : mGlobalRegistrations) aListener.onOreRegistration(tEvent);
	}
	
	/**
	 * This Function adds a global Listener for Recycling Events. These Events let you create Reversal Recipes inside your Machines, and contain all the Components, which can be recycled out of the passed ItemStack.
	 * Unlike the Forge Event System, which replaced Eloraams better System, this Interface ensures, like the great old System, that also all the older Events, which have been called before your Listener-Registration, are getting passed onto your Listener.
	 */
	public void addListener(IOreDictRecyclableListener aListener) {
		if (GAPI.mStartedPostInit) addListenerInternal(aListener); else mBufferedListeners2.add(aListener);
	}
	
	private final Set<IOreDictRecyclableListener> mBufferedListeners2 = new HashSetNoNulls<IOreDictRecyclableListener>();
	private void addListenerInternal(IOreDictRecyclableListener aListener) {
		if (mRecyclableOreDictListeners.add(aListener)) for (OreDictRecyclingContainer tEvent : mRecyclableRegistrations) aListener.onRecycleableRegistration(new OreDictRecyclingContainer(tEvent));
	}
	
	public void onPostLoad() {
		if (GAPI.mStartedPostInit) {
			UT.LoadingBar.start("OreDict", 4);
			UT.LoadingBar.step("Unification Entries");
			registerUnificationEntries();
			UT.LoadingBar.step("Global Listeners");
			for (IOreDictListenerEvent aListener : mBufferedListeners1) addListenerInternal(aListener);
			UT.LoadingBar.step("Recycling Listeners");
			for (IOreDictRecyclableListener aListener : mBufferedListeners2) addListenerInternal(aListener);
			UT.LoadingBar.step("Prefix Listeners");
			for (OreDictPrefix tPrefix : OreDictPrefix.VALUES) tPrefix.onPostLoad();
			UT.LoadingBar.finish();
		}
	}
	
	/**
	 * Used to check if Recipe Outputs accidentally contain uncopied OreDict Items.
	 */
	public boolean isOreDictItem(ItemStack aStack) {
		return mAllRegisteredOres.contains(aStack);
	}
	
	/**
	 * Adds a re-registration for an Ore for with different Name.
	 */
	public boolean addReRegistrationWithReversal(String aOreToReRegister, String aAdditionalName) {
		// Why that Variable you say? Because the OR would kill the second call in its attempt to optimise the return value, and I don't want that second call to be killed.
		boolean rResult = addReRegistration(aOreToReRegister, aAdditionalName);
		return addReRegistration(aAdditionalName, aOreToReRegister) || rResult;
	}
	
	public boolean addAutoBlackListing(String aAutoBlackListedString) {
		return mAutoBlackListedNames.add(aAutoBlackListedString);
	}
	
	public boolean addReRegistration(String aOreToReRegister, String aAdditionalName) {
		mNotUnknownNames.add(aOreToReRegister);
		mNotUnknownNames.add(aAdditionalName);
		if (aOreToReRegister.equals(aAdditionalName)) return F;
		Collection<String> tNames = mReRegistrationMappings.get(aOreToReRegister);
		if (tNames == null) mReRegistrationMappings.put(aOreToReRegister, tNames = new HashSetNoNulls<String>());
		if (tNames.add(aAdditionalName)) {
			// You thought I would not have this Line, to make it possible to add Re-Registrations for already registered things, didn't you?
			for (ItemStack tOreStack : OreDictionary.getOres(aOreToReRegister)) OreDictionary.registerOre(aAdditionalName, tOreStack.copy());
		}
		return T;
	}
	
	/**
	 * Assign Item Data to certain OreDict Strings for recycling and similar. Used for plankWood and slabWood for example.
	 */
	public boolean setAutomaticItemData(String aOreDictName, OreDictItemData aData) {
		mNotUnknownNames.add(aOreDictName);
		mStringToItemDataMappings.put(aOreDictName, aData);
		for (ItemStack tStack : getOres(aOreDictName, F)) {
			addItemData(tStack, OreDictItemData.copy(aData));
		}
		return T;
	}
	
	public OreDictItemData getAutomaticItemData(String aOreDictName) {
		return OreDictItemData.copy(mStringToItemDataMappings.get(aOreDictName));
	}
	
	public Collection<String> getUnknownMaterials() {
		return new ArrayListNoNulls<String>(mUnknownMaterials);
	}
	
	public Collection<String> getUnknownNames() {
		return new ArrayListNoNulls<String>(mUnknownNames);
	}
	
	private static class OreDictEventContainer {
		protected final String mModID;
		protected final OreRegisterEvent mEvent;
		
		protected OreDictEventContainer(String aModID, OreRegisterEvent aEvent) {
			mModID = aModID;
			mEvent = aEvent;
		}
	}
	
	private final List<OreDictEventContainer> mBufferedRegistrations = new ArrayListNoNulls<OreDictEventContainer>();
	
	/** Call from the GT_API Object in order to enable the Prefix/Material processing. */
	public void enableRegistrations() {
		if (GAPI.mStartedInit) {
			UT.LoadingBar.start("OreDict", mBufferedRegistrations.size());
			for (OreDictEventContainer tContainer : mBufferedRegistrations) {
				UT.LoadingBar.step(tContainer.mEvent.Name);
				onOreRegistration2(tContainer.mModID, tContainer.mEvent);
			}
			UT.LoadingBar.finish();
			mBufferedRegistrations.clear();
		}
	}
	
	/** Fluid Containers are not unificatable at all. Also I set Water Bottles to contain 0l of Water. */
	@SubscribeEvent
    public void onFluidContainerRegistration(FluidContainerRegisterEvent aFluidEvent) {
		if (aFluidEvent.data.filledContainer.getItem() == Items.potionitem && UT.Stacks.meta(aFluidEvent.data.filledContainer) == 0) aFluidEvent.data.fluid.amount = 0;
		addToBlacklist(aFluidEvent.data.emptyContainer);
		UT.Fluids.addFluidContainerData(aFluidEvent.data);
    }
	
	@SubscribeEvent
    public void onOreRegistration1(OreRegisterEvent aEvent) {
		ModContainer tContainer = Loader.instance().activeModContainer();
    	String aModID = tContainer==null||mIsRunningInIterationMode?"UNKNOWN":tContainer.getModId();
    	
    	if (aEvent == null || aEvent.Ore == null || aEvent.Ore.getItem() == null || aEvent.Name == null || aEvent.Name.isEmpty()) throw new IllegalArgumentException(aModID + " did something very bad! The registration is too invalid to even be shown properly. This happens only if you register null, invalid Items, empty Strings or even nonexisting Events to the OreDict.");
    	
    	// In order to fix a ThaumCraft Bug I have to ignore this registration under all circumstances. I registered it under the proper Name manually.
    	if (aModID.equals(ModIDs.TC) && aEvent.Name.toLowerCase().endsWith("uicksilver")) return;
    	
//    	GT_Log.ore.println(aModID + " -> " + aEvent.Name);
    	
		aEvent.Ore.stackSize = 1;
		
		mAllRegisteredOres.add(aEvent.Ore);
		
		if (aEvent.Name.contains("|") || aEvent.Name.contains("*") || aEvent.Name.contains(":") || aEvent.Name.contains(".") || aEvent.Name.contains("$")) return;
		
		if (aEvent.Name.contains(" ")) {
			OreDictionary.registerOre(aEvent.Name.replaceAll(" ", ""), aEvent.Ore.copy());
			return;
		}
		
		Collection<String> tReRegistrations = mReRegistrationMappings.get(aEvent.Name);
		if (tReRegistrations != null) for (String tName : tReRegistrations) OreDictionary.registerOre(tName, aEvent.Ore.copy());
		if (mAutoBlackListedNames.contains(aEvent.Name)) addToBlacklist(aEvent.Ore);
		
		if (!mIgnoredNames.contains(aEvent.Name)) if (GAPI.mStartedInit) onOreRegistration2(aModID, aEvent); else mBufferedRegistrations.add(new OreDictEventContainer(aModID, aEvent));
	}
	
	public void onOreRegistration2(String aModID, OreRegisterEvent aEvent) {
		OreDictPrefix aPrefix = null;
		OreDictMaterial aMaterial = null;
		
		if (aEvent.Ore.getItem() instanceof IOreDictOptimizedParsingItem) {
			aPrefix = ((IOreDictOptimizedParsingItem)aEvent.Ore.getItem()).getPrefix(UT.Stacks.meta(aEvent.Ore));
			if (!aEvent.Name.startsWith(aPrefix.mNameInternal)) aPrefix = null;
		}
		
		if (aPrefix == null) aPrefix = OreDictPrefix.get(aEvent.Name);
		
		addItemData(aEvent.Ore, OreDictItemData.copy(mStringToItemDataMappings.get(aEvent.Name)));
		
    	boolean aNotAlreadyRegisteredName = mAlreadyRegisteredNames.add(aEvent.Name);
    	
		if (aPrefix == null) {
			if (!mNotUnknownNames.contains(aEvent.Name)) mUnknownNames.add(aEvent.Name);
		} else {
			if (aPrefix != aPrefix.mTargetRegistration) {
				OreDictionary.registerOre(aEvent.Name.replaceFirst(aPrefix.mNameInternal, aPrefix.mTargetRegistration.mNameInternal), aEvent.Ore.copy());
				return;
			}
			
			if (aPrefix.contains(TD.Prefix.UNIFICATABLE_RECIPES)) addToBlacklist(aEvent.Ore);
			
			String tName = aEvent.Name.replaceFirst(aPrefix.mNameInternal, "");
			
			if (tName.length() > 0) {
				// This is a vanilla Forge Rule. I just make sure it gets applied. All Dyes no matter of which Colour have to be registered under the Name "dye" too.
				if (aPrefix == OP.dye) OreDictionary.registerOre("dye", aEvent.Ore.copy());
				// This is a GregTech Rule. GT Gears are convertible to BC Type Gears, but not the other way around, to prevent exploits with the Reverse Recipes.
				if (aPrefix == OP.gearGt) OreDictionary.registerOre(OP.gear.mNameInternal + tName, aEvent.Ore.copy());
				
				aMaterial = OreDictMaterial.MATERIAL_MAP.get(tName);
				
				if (aPrefix.contains(TD.Prefix.MATERIAL_BASED)) {
					if (aMaterial == null) aMaterial = OreDictMaterial.createMaterial(-1, tName, tName).add(TD.Properties.INVALID_MATERIAL, TD.Properties.AUTO_BLACKLIST, TD.Properties.AUTO_MATERIAL);
					if (aMaterial.contains(TD.Properties.AUTO_MATERIAL) && !mNotUnknownNames.contains(aEvent.Name)) mUnknownMaterials.add(aEvent.Name);
					if (aMaterial != aMaterial.mTargetRegistration) OreDictionary.registerOre(aPrefix.mNameInternal + aMaterial.mTargetRegistration.mNameInternal, aEvent.Ore.copy());
					if (aMaterial.contains(TD.Properties.AUTO_BLACKLIST)) addToBlacklist(aEvent.Ore);
					for (OreDictMaterial tReRegisteredMaterial : aMaterial.mReRegistrations) OreDictionary.registerOre(aPrefix.mNameInternal + tReRegisteredMaterial.mNameInternal, aEvent.Ore.copy());
					if (!aMaterial.contains(TD.Properties.INVALID_MATERIAL) && aPrefix.contains(TD.Prefix.UNIFICATABLE)) setTarget(aPrefix, aMaterial, aEvent.Ore, F, T);
				}
			}
			
			aPrefix.onOreRegistration(new OreDictRegistrationContainer(aPrefix, aMaterial, aEvent.Name, aEvent.Ore.copy(), aEvent, aModID, aNotAlreadyRegisteredName));
		}
		
		OreDictRegistrationContainer tRegistration = new OreDictRegistrationContainer(aPrefix, aMaterial, aEvent.Name, aEvent.Ore.copy(), aEvent, aModID, aNotAlreadyRegisteredName);
		
		for (IOreDictListenerEvent tListener : mGlobalOreDictListeners) tListener.onOreRegistration(tRegistration);
		mGlobalRegistrations.add(tRegistration);
	}
	
	// ================================================================ //
	// A lot of the Code was copy-pasted from the old OreDictUnificator //
	// ================================================================ //
	
	private final Map<String, ItemStack> sName2StackMap = new HashMap<String, ItemStack>();
	private final Map<ItemStackContainer, OreDictItemData> sItemStack2DataMap = new ItemStackMap<ItemStackContainer, OreDictItemData>();
	private final ItemStackSet<ItemStackContainer> sNoUnificationList = new ItemStackSet<ItemStackContainer>();
	private int isRegisteringOre = 0, isAddingOre = 0;
	
	/**
	 * The Blacklist just prevents the Item from being unificated into something else.
	 * Useful if you have things like the Industrial Diamond, which is better than regular Diamond, but also usable in absolutely all Diamond Recipes.
	 */
	public void addToBlacklist(ItemStack aStack) {
		if (UT.Stacks.valid(aStack) && sNoUnificationList.add(new ItemStackContainer(aStack))) {
			OreDictItemData tData = getItemData(aStack);
			if (tData != null) tData.mBlackListed = tData.mBlocked = T;
		}
	}
	
	public boolean isBlacklisted(ItemStack aStack) {
		return sNoUnificationList.contains(aStack, T);
	}
	
	public boolean addTarget(OreDictPrefix aPrefix, OreDictMaterial aMaterial, ItemStack aStack) {
		return setTarget(aPrefix, aMaterial, aStack, F, F);
	}
	
	public boolean setTarget(OreDictPrefix aPrefix, OreDictMaterial aMaterial, ItemStack aStack) {
		return setTarget(aPrefix, aMaterial, aStack, T, F);
	}
	
	public boolean setTarget(OreDictPrefix aPrefix, OreDictMaterial aMaterial, ItemStack aStack, boolean aOverwrite, boolean aAlreadyRegistered) {
		if (aMaterial == null || aPrefix == null || UT.Stacks.invalid(aStack) || UT.Stacks.meta(aStack) == W) return F;
		isAddingOre++;
		(aStack=UT.Stacks.amount(1, aStack)).setTagCompound(null);
		if (!aAlreadyRegistered) registerOre(aPrefix.mNameInternal + aMaterial.mNameInternal, aStack);
		addAssociation(aPrefix, aMaterial, aStack);
		if (!isBlacklisted(aStack) && (aOverwrite || UT.Stacks.invalid(sName2StackMap.get(aPrefix.mNameInternal + aMaterial.mNameInternal)))) sName2StackMap.put(aPrefix.mNameInternal + aMaterial.mNameInternal, aStack);
		isAddingOre--;
		return T;
	}
	
	public ItemStack getFirstOre(Object aName, long aAmount) {
		if (UT.Code.stringInvalid(aName)) return null;
		ItemStack tStack = sName2StackMap.get(aName.toString());
		if (UT.Stacks.valid(tStack)) return UT.Stacks.amount(aAmount, tStack);
		return UT.Stacks.amount(aAmount, getOres(aName, F).toArray());
	}
	
	public ItemStack getStack(Object aName, long aAmount) {
		return getStack(aName, null, aAmount, T, T);
	}
	
	public ItemStack getStack(Object aName, ItemStack aReplacement, long aAmount) {
		return getStack(aName, aReplacement, aAmount, T, T);
	}
	
	public ItemStack getStack(OreDictPrefix aPrefix, OreDictMaterial aMaterial, long aAmount) {
		return getStack(aPrefix, aMaterial, null, aAmount);
	}
	
	public ItemStack getStack(OreDictPrefix aPrefix, OreDictMaterial aMaterial, ItemStack aReplacement, long aAmount) {
		return getStack(aPrefix.mNameInternal + aMaterial.mNameInternal, aReplacement, aAmount, F, T);
	}
	
	public ItemStack getStack(Object aName, ItemStack aReplacement, long aAmount, boolean aMentionPossibleTypos, boolean aNoInvalidAmounts) {
		if (aNoInvalidAmounts && aAmount < 1) return null;
		if (!sName2StackMap.containsKey(aName.toString()) && aMentionPossibleTypos) FMLLog.info("Unknown Key for Unification, Typo? " + aName);
		return UT.Stacks.amount(aAmount, sName2StackMap.get(aName.toString()), getFirstOre(aName, aAmount), aReplacement);
	}
	
	public ItemStack[] setStackArray(boolean aUseBlackList, ItemStack... aStacks) {
		for (int i = 0; i < aStacks.length; i++) aStacks[i] = getStack(aUseBlackList, UT.Stacks.copy(aStacks[i]));
		return aStacks;
	}
	
	public ItemStack[] getStackArray(boolean aUseBlackList, Object... aStacks) {
		ItemStack[] rStacks = new ItemStack[aStacks.length];
		for (int i = 0; i < aStacks.length; i++) rStacks[i] = getStack(aUseBlackList, UT.Stacks.copy(aStacks[i]));
		return rStacks;
	}
	
	public ItemStack setStack(boolean aUseBlackList, ItemStack aStack) {
		if (UT.Stacks.invalid(aStack)) return aStack;
		ItemStack tStack = getStack(aUseBlackList, aStack);
		if (UT.Stacks.equal(aStack, tStack)) return aStack;
		aStack.func_150996_a(tStack.getItem());
		return UT.Stacks.meta(aStack, UT.Stacks.meta(tStack));
	}
	
	public ItemStack getStack(boolean aUseBlackList, ItemStack aStack) {
		if (UT.Stacks.invalid(aStack)) return null;
		OreDictItemData tAssociation = getAssociation(aStack, F);
		ItemStack rStack = null;
		if (tAssociation == null || (aUseBlackList && tAssociation.mBlocked)) return UT.Stacks.copy(aStack);
		if (tAssociation.mUnificationTarget == null) tAssociation.mUnificationTarget = sName2StackMap.get(tAssociation.toString());
		if (UT.Stacks.invalid(rStack = UT.Stacks.copy(tAssociation.mUnificationTarget))) return UT.Stacks.copy(aStack);
		assert rStack != null;
		rStack.setTagCompound(aStack.getTagCompound());
		return UT.Stacks.amount(aStack.stackSize, rStack);
	}
	
	public boolean addItemData(ItemStack aStack, OreDictItemData aData) {
		if (getItemData(aStack) == null) return setItemData(aStack, aData);
		return F;
	}
	
	public boolean setItemData(ItemStack aStack, OreDictItemData aData) {
		if (UT.Stacks.invalid(aStack) || aData == null || getAssociation(aStack, F) != null) return F;
		if (aStack.stackSize > 1) {
			if (aData.mMaterial != null) aData.mMaterial.mAmount /= aStack.stackSize;
			for (OreDictMaterialStack tMaterial : aData.mByProducts) tMaterial.mAmount /= aStack.stackSize;
			aStack = UT.Stacks.amount(1, aStack);
		}
		if (!aData.mBlackListed) aData.mBlackListed = isBlacklisted(aStack);
		if (!aData.mBlocked) aData.mBlocked = (aData.mBlackListed || UT.Stacks.block(aStack) != Blocks.air || UT.Fluids.getFluidForFilledItem(aStack, true) != null || (aStack.getItem() instanceof IFluidContainerItem && ((IFluidContainerItem)aStack.getItem()).getCapacity(aStack) > 0));
		sItemStack2DataMap.put(new ItemStackContainer(aStack), aData);
		if (aData.hasValidMaterialData()) {
			long tValidMaterialAmount = aData.mMaterial.mMaterial.contains(TD.Processing.UNRECYCLABLE)?0:aData.mMaterial.mAmount>=0?aData.mMaterial.mAmount:U;
			for (OreDictMaterialStack tMaterial : aData.mByProducts) tValidMaterialAmount += tMaterial.mMaterial.contains(TD.Processing.UNRECYCLABLE)?0:tMaterial.mAmount>=0?tMaterial.mAmount:U;
			if (tValidMaterialAmount < U) GT_ModHandler.addToRecyclerBlackList(aStack);
		}
		if (!aData.hasValidPrefixData() || aData.mPrefix.contains(TD.Prefix.RECYCLABLE)) {
			OreDictRecyclingContainer tRegistration = new OreDictRecyclingContainer(aStack, aData);
			for (IOreDictRecyclableListener tListener : mRecyclableOreDictListeners) tListener.onRecycleableRegistration(tRegistration);
			mRecyclableRegistrations.add(tRegistration);
		}
		return true;
	}
	
	public boolean addAssociation(OreDictPrefix aPrefix, OreDictMaterial aMaterial, ItemStack aStack) {
		if (aPrefix == null || aMaterial == null || UT.Stacks.invalid(aStack)) return false;
		if (UT.Stacks.meta(aStack) == W) for (byte i = 0; i < 16; i++) setItemData(UT.Stacks.copyAmountAndMeta(1, i, aStack), new OreDictItemData(aPrefix, aMaterial));
		return setItemData(aStack, new OreDictItemData(aPrefix, aMaterial));
	}
	
	public OreDictItemData getItemData(ItemStack aStack) {
		return getItemData(aStack, false);
	}
	
	public OreDictItemData getItemData(ItemStack aStack, boolean aAllowOverride) {
		if (UT.Stacks.invalid(aStack)) return null;
		if (aAllowOverride) {
			if (aStack.getItem() instanceof IOreDictItemDataOverrideItem) {
				OreDictItemData rData = ((IOreDictItemDataOverrideItem)aStack.getItem()).getOreDictItemData(aStack);
				if (rData != null) return rData;
			}
			NBTTagCompound tNBT = aStack.getTagCompound();
			if (tNBT != null && tNBT.hasKey(NBT_RECYCLING_MATS)) {
				List<OreDictMaterialStack> tList = OreDictMaterialStack.loadList(NBT_RECYCLING_MATS, tNBT);
				if (!tList.isEmpty()) {
					OreDictMaterialStack tMain = tList.get(0);
					tList.remove(0);
					return new OreDictItemData(tMain, tList.toArray(ZL_MATERIALSTACK));
				}
			}
		}
		OreDictItemData rData = sItemStack2DataMap.get(new ItemStackContainer(aStack));
		if (rData != null) return rData;
		rData = sItemStack2DataMap.get(new ItemStackContainer(UT.Stacks.copyMeta(W, aStack)));
		if (rData != null) return rData;
		if (aAllowOverride && aStack.getItem().isDamageable()) {
			rData = sItemStack2DataMap.get(new ItemStackContainer(UT.Stacks.copyMeta(0, aStack)));
			if (rData != null && rData.mUseVanillaDamage) {
				OreDictMaterialStack[] tByProducts = new OreDictMaterialStack[rData.mByProducts.length];
				for (int i = 0; i < tByProducts.length; i++) tByProducts[i] = OM.stack(rData.mByProducts[i].mMaterial, UT.Code.units(aStack.getMaxDamage()-aStack.getItemDamage(), aStack.getMaxDamage(), rData.mByProducts[i].mAmount, F));
				rData = new OreDictItemData(OM.stack(rData.mMaterial.mMaterial, UT.Code.units(aStack.getMaxDamage()-aStack.getItemDamage(), aStack.getMaxDamage(), rData.mMaterial.mAmount, F)), tByProducts);
			}
		}
		return rData;
	}
	
	public OreDictItemData getAssociation(ItemStack aStack, boolean aOverwrite) {
		OreDictItemData rData = getItemData(aStack, aOverwrite);
		return rData != null && rData.hasValidPrefixMaterialData() ? rData : null;
	}
	
    public boolean registerOre(OreDictPrefix aPrefix, OreDictMaterial aMaterial, ItemStack aStack) {
    	return registerOre(aPrefix.mNameInternal + aMaterial.mNameInternal, aStack);
    }
    
    public boolean registerOre(Object aName, ItemStack aStack) {
    	if (aName == null || UT.Stacks.invalid(aStack)) return F;
    	if (GAPI.mStartedPostInit) throw new IllegalStateException("Late OreDict Registration using GT OreDict Utility. Only @Init and @PreInit are allowed for this when you use this Function instead of the Forge one.");
    	String tName = aName.toString();
    	if (UT.Code.stringInvalid(tName)) return F;
    	ArrayList<ItemStack> tList = getOres(tName, F);
    	for (int i = 0; i < tList.size(); i++) if (UT.Stacks.equal(tList.get(i), aStack, T)) return F;
    	isRegisteringOre++;
    	OreDictionary.registerOre(tName, UT.Stacks.amount(1, aStack));
    	isRegisteringOre--;
    	return T;
    }
    
    public boolean isRegisteringOres() {
    	return isRegisteringOre > 0;
    }
    
    public boolean isAddingOres() {
    	return isAddingOre > 0;
    }
    
    private void registerUnificationEntries() {
    	for (OreDictItemData tPrefixMaterial : sItemStack2DataMap.values()) tPrefixMaterial.mUnificationTarget = null;
		for (OreDictRegistrationContainer tEvent : mGlobalRegistrations) {
			if (tEvent.mPrefix != null && tEvent.mMaterial != null && tEvent.mPrefix.contains(TD.Prefix.UNIFICATABLE) && !tEvent.mMaterial.contains(TD.Properties.INVALID_MATERIAL)) {
				addAssociation(tEvent.mPrefix, tEvent.mMaterial, tEvent.mStack);
				setTarget(tEvent.mPrefix, tEvent.mMaterial, tEvent.mStack, mUnificationConfig.get("specialunificationtargets." + tEvent.mModName, tEvent.mEvent.Name, F), T);
			}
		}
    }
    
	public static boolean isItemStackInstanceOf(ItemStack aStack, Object aName) {
		if (UT.Code.stringInvalid(aName) || UT.Stacks.invalid(aStack)) return F;
		for (ItemStack tOreStack : getOres(aName.toString(), F)) if (UT.Stacks.equal(tOreStack, aStack, T)) return T;
		return F;
	}
	
    /** @return a Copy of the OreDictionary.getOres() List */
    public static ArrayList<ItemStack> getOres(OreDictPrefix aPrefix, OreDictMaterial aMaterial, boolean aTransformWildcardBlocksTo16) {
    	return getOres(aPrefix.mNameInternal + aMaterial.mNameInternal, aTransformWildcardBlocksTo16);
    }
    
    /** @return a Copy of the OreDictionary.getOres() List */
    public static ArrayList<ItemStack> getOres(Object aOreName, boolean aTransformWildcardBlocksTo16) {
    	String aName = aOreName==null?"":aOreName.toString();
    	ArrayList<ItemStack> rList = new ArrayListNoNulls<ItemStack>(), tList;
    	if (UT.Code.stringValid(aName)) {
    		if (aTransformWildcardBlocksTo16) {
        		tList = OreDictionary.getOres(aName);
        		for (ItemStack tStack : tList) {
        			if (UT.Stacks.meta(tStack) == W && UT.Stacks.block(tStack) != Blocks.air) {
        				for (int i = 0; i < 16; i++) rList.add(UT.Stacks.make(tStack.getItem(), 1, i));
        			} else {
        				rList.add(tStack);
        			}
        		}
    		} else {
    			rList.addAll(OreDictionary.getOres(aName));
    		}
    	}
    	return rList;
    }
}