package gregapi.tileentity.base;

import static gregapi.data.CS.*;
import gregapi.block.multitileentity.IMultiTileEntity.IMTE_BreakBlock;
import gregapi.block.multitileentity.IMultiTileEntity.IMTE_OnBlockExploded;
import gregapi.block.multitileentity.MultiTileEntityRegistry;
import gregapi.util.OM;
import gregapi.util.UT;
import net.minecraft.entity.item.EntityItem;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.inventory.IInventory;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTTagList;
import net.minecraft.world.Explosion;

/**
 * @author Gregorius Techneticies
 */
public abstract class TileEntityBase05Inventories extends TileEntityBase04MultiTileEntities implements IInventory, IMTE_OnBlockExploded, IMTE_BreakBlock {
	private ItemStack[] mInventory = ZL_ITEMSTACK;
	
	public boolean mInventoryChanged = F;
	
    @Override
	public void readFromNBT2(NBTTagCompound aNBT) {
        // Standard readFromNBT process to load Inventory.
        mInventory = getDefaultInventory(aNBT);
        NBTTagList tList = aNBT.getTagList(NBT_INV_LIST, 10);
        for (int i = 0; i < tList.tagCount(); i++) {
            NBTTagCompound tNBT = tList.getCompoundTagAt(i);
            int tSlot = tNBT.getShort("s");
            if (tSlot >= 0 && tSlot < mInventory.length) mInventory[tSlot] = UT.Stacks.load(tNBT);
        }
    }
    
    @Override
	public void writeToNBT2(NBTTagCompound aNBT) {
    	if (mInventory != null && mInventory.length > 0) {
	        NBTTagList tList = new NBTTagList();
	        for (short tSlot = 0; tSlot < mInventory.length; tSlot++) if (mInventory[tSlot] != null) {
	            NBTTagCompound tNBT = UT.NBT.getNBTShort(null, "s", tSlot);
	            mInventory[tSlot].writeToNBT(tNBT);
	            tList.appendTag(tNBT);
	        }
	        aNBT.setTag(NBT_INV_LIST, tList);
    	}
    }
    
    public ItemStack[] getDefaultInventory(NBTTagCompound aNBT) {
    	short tSize = aNBT.getShort(NBT_INV_SIZE);
    	return tSize > 0 ? new ItemStack[tSize] : ZL_ITEMSTACK;
    }
    
    @Override
    public void onTickResetChecks(long aTimer, boolean aIsServerSide) {
    	super.onTickResetChecks(aTimer, aIsServerSide);
    	mInventoryChanged = F;
    }
    
    public final void slot(int aIndex, ItemStack aStack) {mInventory[aIndex] = aStack;}
    public final ItemStack slot(int aIndex) {return mInventory[aIndex];}
    
    @Override protected void updateInventory() {mInventoryChanged = T;}
	@Override public boolean isUseableByPlayer(EntityPlayer aPlayer) {return aPlayer.getDistanceSq(xCoord + 0.5D, yCoord + 0.5D, zCoord + 0.5D) <= 64D;}
	@Override public void openInventory() {/**/}
	@Override public void closeInventory() {/**/}
	@Override public int getInventoryStackLimit() {return 64;}
	@Override public void onExploded(Explosion aExplosion) {for (int i = 0; i < mInventory.length; i++) if (worldObj.rand.nextInt(3) != 0) mInventory[i] = null; worldObj.setBlockToAir(xCoord, yCoord, zCoord);}
	@Override public void markDirty() {super.markDirty(); mInventoryChanged = T;}
    @Override public ItemStack decrStackSize(int aSlot, int aDecrement) {mInventoryChanged = T; if (mInventory[aSlot] != null) {if (mInventory[aSlot].stackSize <= aDecrement) {ItemStack tStack = mInventory[aSlot]; mInventory[aSlot] = null; return tStack;} ItemStack rStack = mInventory[aSlot].splitStack(aDecrement); if (mInventory[aSlot].stackSize <= 0) mInventory[aSlot] = null; return rStack;} return null;}
	@Override public ItemStack getStackInSlotOnClosing(int aSlot) {ItemStack rStack = mInventory[aSlot]; mInventory[aSlot] = null; return rStack;}
    @Override public ItemStack getStackInSlot(int aSlot) {return mInventory[aSlot];}
	@Override public String getInventoryName() {String rName = getCustomName(); if (UT.Code.stringValid(rName)) return rName; MultiTileEntityRegistry tRegistry = MultiTileEntityRegistry.getRegistry(getMultiTileEntityRegistryID()); return tRegistry==null?getClass().getName():tRegistry.getLocal(getMultiTileEntityID());}
	@Override public int getSizeInventory() {return mInventory==null?0:mInventory.length;}
	@Override public void setInventorySlotContents(int aSlot, ItemStack aStack) {mInventoryChanged = T; mInventory[aSlot] = OM.get(aStack);}
	@Override public boolean hasCustomInventoryName() {return getCustomName() != null;}
	@Override public boolean isItemValidForSlot(int aSlot, ItemStack aStack) {return T;}
	@Override public boolean breakBlock() {if (isServerSide()) for (short i = 0; i < mInventory.length; i++) if (mInventory[i] != null && canDrop(i)) worldObj.spawnEntityInWorld(new EntityItem(worldObj, xCoord+0.5, yCoord+0.5, zCoord+0.5, UT.Stacks.copy(mInventory[i]))); return F;}
	public ItemStack[] getInventory() {return mInventory;}
	public void setInventory(ItemStack[] aInventory) {mInventory = aInventory;}
	public void removeAllDroppableNullStacks() {for (int i = 0; i < mInventory.length; i++) if (canDrop(i) && mInventory[i] != null && mInventory[i].stackSize <= 0) mInventory[i] = null;}
	
	public abstract boolean canDrop(int aInventorySlot);
	
	public boolean addStackToSlot(int aIndex, ItemStack aStack) {
		if (UT.Stacks.invalid(aStack)) return T;
		if (aIndex < 0 || aIndex >= getSizeInventory()) return F;
		ItemStack tStack = getStackInSlot(aIndex);
		if (UT.Stacks.invalid(tStack)) {
			setInventorySlotContents(aIndex, aStack);
			return T;
		}
		aStack = OM.get(aStack);
		if (UT.Stacks.equal(tStack, aStack) && tStack.stackSize + aStack.stackSize <= Math.min(tStack.getMaxStackSize(), getInventoryStackLimit())) {
			tStack.stackSize+=aStack.stackSize;
			mInventoryChanged = T;
			return T;
		}
		return F;
	}
}